import datetime
import json


def handler_(obj):
    if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
        return str(obj)


def json_dumps(data):
    return json.dumps(data, default=handler_)


def json_loads(data):
    return json.loads(data)


import tornado.web
import tornado.gen

from . import utils


class BaseApiHandler(tornado.web.RequestHandler):

    def response(self, data, code=200):
        if code != 200:
            # if something went wrong, we include returned HTTP code in the
            # JSON response
            data['status'] = code

        self.write(utils.json_dumps(data) + "\n")
        self.set_status(code)

    def error(self, message, code=500):
        self.response({'message': message}, code)

    def options(self, *args, **kwargs):
        pass


class StatsHandler(BaseApiHandler):

    @tornado.gen.coroutine
    def get(self, *args, **kwargs):
        stats = yield self.application.monitor.get_stats()
        self.response({'stats': stats}, 200)


class UrlsListHandler(BaseApiHandler):

    @tornado.gen.coroutine
    def get(self, *args, **kwargs):
        data = yield self.application.monitor.iter_urls()
        objects = [
            dict(url=url, response=response)
            for url, response
            in data
        ]
        self.response({'objects': objects}, 200)


class UrlDetailsHandler(BaseApiHandler):

    @tornado.gen.coroutine
    def get(self, *args, **kwargs):
        url = self.get_argument('url', None)

        try:
            info = yield self.application.monitor.get_info(url)
            self.response({'response': info}, 200)
        except KeyError:
            error = (
                u"Url: {} is not monitored. Perhaps slash is missing"
                u" or is redundant."
            ).format(url)
            self.response({'error': error}, 400)


import abc
from itertools import izip
from datetime import datetime

import tornado.gen

from tormon.api import utils


DEFAULT_TIMESTAMP = 0.0
DEFAULT_STATUS_CODE = None
DEFAULT_HEADER_SET = {}


class IBaseWriter(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def iter_data(self):
        pass

    @abc.abstractmethod
    def write_response(self, url, response):
        pass

    @abc.abstractmethod
    def write_error(self, url, error):
        pass

    @abc.abstractmethod
    def get_info(self, url):
        pass

    @abc.abstractmethod
    def get_stats(self):
        pass

    @property
    def updated_at(self):
        return datetime.now()

    def get_response_data(self, response):
        return {
            u'code': response.code,
            u'time': response.request_time,
            u'headers': response.headers,
            u'updated_at': self.updated_at
        }

    def get_error_data(self, error):
        error_data = error.__dict__

        return {
            u'code': error_data.get(u'code', DEFAULT_STATUS_CODE),
            u'headers': error_data.get(u'headers', DEFAULT_HEADER_SET),
            u'time': error_data.get(u'request_time', DEFAULT_TIMESTAMP),
            u'updated_at': self.updated_at,
            u'error': str(error)
        }


class MemoryWriter(IBaseWriter):

    def __init__(self, *args, **kwargs):
        self.url_status = {}

    def write_response(self, url, response):
        data = self.get_response_data(response=response)
        self.url_status[url] = data

    def write_error(self, url, error):
        data = self.get_error_data(error=error)
        self.url_status[url] = data

    @tornado.gen.coroutine
    def iter_data(self):
        data = self.url_status
        raise tornado.gen.Return(data.iteritems())

    @tornado.gen.coroutine
    def get_info(self, url):
        data = self.url_status[url]
        raise tornado.gen.Return(data)

    @tornado.gen.coroutine
    def get_stats(self):
        data = yield tornado.gen.Task(self.iter_data)
        urls = [
            url for url, _ in data
        ]

        stats = {
            u'total_monitored': len(urls),
            u'urls': urls
        }
        raise tornado.gen.Return(stats)


class RedisWriter(IBaseWriter):
    KEY_PREFIX = u'tormon-'

    def __init__(self, *args, **kwargs):
        import redis
        self.r = redis.StrictRedis()

    def key(self, base):
        return u"{}{}".format(self.KEY_PREFIX, base)

    @tornado.gen.coroutine
    def iter_data(self):
        match = u"{}*".format(self.KEY_PREFIX)
        count = None
        results = {}

        keys = [
            key for
            key in self.r.scan_iter(match=match, count=count)
        ]

        values = self.r.mget(keys)

        keys = map(lambda x: x.replace(self.KEY_PREFIX, u''), keys)
        values = map(lambda x: utils.json_loads(x), values)
        results.update(dict(izip(keys, values)))

        raise tornado.gen.Return(results.iteritems())

    @tornado.gen.coroutine
    def write_response(self, url, response):
        data = self.get_response_data(response=response)
        json_data = utils.json_dumps(data)
        self.r.set(self.key(url), json_data)

    @tornado.gen.coroutine
    def write_error(self, url, error):
        data = self.get_error_data(error=error)
        json_data = utils.json_dumps(data)
        self.r.set(self.key(url), json_data)

    @tornado.gen.coroutine
    def get_info(self, url):
        key = self.key(url)
        data = self.r.get(key)
        if not data:
            raise KeyError(u"Key: {} does not exist in db".format(key))
        raise tornado.gen.Return(utils.json_loads(data))

    @tornado.gen.coroutine
    def get_stats(self):
        data = yield tornado.gen.Task(self.iter_data)
        urls = [
            url for url, _ in data
        ]

        stats = {
            u'total_monitored': len(urls),
            u'urls': urls
        }
        raise tornado.gen.Return(stats)


from setuptools import setup, find_packages

requires = [
    'pycurl==7.19.5.1',
    'tornado==4.2.1'
]

test_requires = [
]


setup(
    name='tormon',
    install_requires=requires,
    tests_require=test_requires,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    version='0.0.1',
    description=(
        'Tornado based application monitor',
    ),
    author='jnosal',
    author_email='jacek.nosal@outlook.com',
    url='https://github.com/jnosal/tormon',
    keywords=[
        'tornado', 'redis', 'tormon'
    ],
    entry_points={
        'console_scripts': [
            'tor-mon = tormon.run:main',
        ],
    },
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
